#if defined(_WIN32)
#define _CRT_SECURE_NO_DEPRECATE
#endif

/*---------------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
#include <time.h>

#include "ini.h"
#include "libuism.h"

#if defined(__linux__) || defined(__QNX__)
    #include <pthread.h>
    #include <signal.h>
    #include <unistd.h>
    #include <fcntl.h>
    #include <stdarg.h>
#elif defined(_WIN32)
    #include <process.h>
    #include <windows.h>
#endif

/*---------------------------------------------------------------------------------*/
#define FILE_PATH_MAX   1024
#define SETUP_INI       "setup.ini"
#define TYPE_EXIT       3

/*---------------------------------------------------------------------------------*/
typedef struct
{
    /* [setting] */
    int  debug_level;

    /* [monitor] */
    int  interval;
    bool auto_scan;
    bool check_status;

    /* [sensor] */
    bool HCHO;
    bool PM;
    bool CO2;
    bool CO;
    bool O3;
    bool TVOC;
    bool TH;

} INI_ITEM;

/*---------------------------------------------------------------------------------*/
static INI_ITEM ini_item =
{
    .debug_level  = DBG_LVL_NONE,
    .interval     = 1000,
    .check_status = false
};

static bool isRunning = false;

/*---------------------------------------------------------------------------------*/
void data_cb_func (int error_code, int name, DATA_INFO *info)
{
    int i, temp, RH;

    if(error_code != UISM_SUCCESS)
        printf("error_code: %d, name: %X\n", error_code, name);
    else
    {
        switch(name)
        {
            case SENSOR_NAME_HCHO:
                printf("HCHO  [ unit: 0.01    ppm ] : %-5d\n", ((info->data1[0] << 8) + info->data1[1]));
                break;

            case SENSOR_NAME_PM:
                printf("PM2.5 [ unit: 0.1  ug/m^3 ] : %-5d\n", ((info->data1[0] << 8) + info->data1[1]));
                printf("PM10  [ unit: 0.1  ug/m^3 ] : %-5d\n", ((info->data2[0] << 8) + info->data2[1]));
                break;

            case SENSOR_NAME_CO2:
                printf("CO2   [ unit: 1       ppm ] : %-5d",   ((info->data1[0] << 8) + info->data1[1]));
                if(ini_item.check_status)
                    printf(" (status: %d)", info->status);
                printf("\n");
                break;

            case SENSOR_NAME_CO:
                printf("CO    [ unit: 0.1     ppm ] : %-5d",   ((info->data1[0] << 8) + info->data1[1]));
                if(ini_item.check_status)
                    printf(" (status: %d)", info->status);
                printf("\n");
                break;

            case SENSOR_NAME_O3:
                printf("O3    [ unit: 0.01    ppm ] : %-5d",   ((info->data1[0] << 8) + info->data1[1]));
                if(ini_item.check_status)
                    printf(" (status: %d)", info->status);
                printf("\n");
                break;

            case SENSOR_NAME_TVOC:
                printf("TVOC  [ unit: 0.001   ppm ] : %-5d",   ((info->data1[0] << 8) + info->data1[1]));
                if(ini_item.check_status)
                    printf(" (status: %d)", info->status);
                printf("\n");
                break;

            case SENSOR_NAME_TH:
                temp = (-45 * 100) + (175 * ((info->data1[0] << 8) + info->data1[1]) * 100) / 0xFFFF;
                RH   = ( -6 * 100) + (125 * ((info->data1[2] << 8) + info->data1[3]) * 100) / 0xFFFF;

                RH   = (RH > 100 * 100) ? (100 * 100) : RH;
                RH   = (RH <         0) ? (        0) : RH;

                printf("TEMP  [ unit: 0.01      C ] : %-5d\n",  temp);
                printf("RH    [ unit: 0.01      %% ] : %-5d\n", RH);
                break;

            default:
                printf("name: %d, buf:", name);
                for(i=0; i<info->data1_len; i++)
                    printf(" [%02X]", info->data1[i]);
                printf("\n");
                for(i=0; i<info->data2_len; i++)
                    printf(" [%02X]", info->data2[i]);
                printf("\n");
                break;
        }
    }

} /* END: data_cb_func() */

/*---------------------------------------------------------------------------------*/
static int handler_ini (void *user, const char *section, const char *name, const char *value)
{
    #define MATCH(s, n)     strcmp(section, s) == 0 && strcmp(name, n) == 0

    INI_ITEM *p_ini = (INI_ITEM *)user;

    /* [setting] section */
    if(MATCH("setting", "debug_level"))
    {
        p_ini->debug_level = (int) atoi(value);
    }

    /* [monitor] section */
    else if(MATCH("monitor", "interval"))
    {
        p_ini->interval = (int) atoi(value);
    }
    else if(MATCH("monitor", "auto_scan"))
    {
        p_ini->auto_scan = (bool) atoi(value);
    }
    else if(MATCH("monitor", "check_status"))
    {
        p_ini->check_status = (bool) atoi(value);
    }

    /* [sensor] section */
    else if(MATCH("sensor", "HCHO"))
    {
        p_ini->HCHO = (bool) atoi(value);
    }
    else if(MATCH("sensor", "PM"))
    {
        p_ini->PM = (bool) atoi(value);
    }
    else if(MATCH("sensor", "CO2"))
    {
        p_ini->CO2 = (bool) atoi(value);
    }
    else if(MATCH("sensor", "CO"))
    {
        p_ini->CO = (bool) atoi(value);
    }
    else if(MATCH("sensor", "O3"))
    {
        p_ini->O3 = (bool) atoi(value);
    }
    else if(MATCH("sensor", "TVOC"))
    {
        p_ini->TVOC = (bool) atoi(value);
    }
    else if(MATCH("sensor", "TH"))
    {
        p_ini->TH = (bool) atoi(value);
    }

    /* unknown section/name, error */
    else
    {
        return 0;
    }

    return 1;

} /* END: handler_ini() */

/*---------------------------------------------------------------------------------*/
static void get_ini_setting (void)
{
    char dest[FILE_PATH_MAX];

    memset(dest, 0, sizeof(dest));

    #if defined(__linux__)
        char path[FILE_PATH_MAX];
        int  len, i;

        memset(path, 0, sizeof(path));

        pid_t pid = getpid();

        sprintf(path, "/proc/%d/exe", pid);
        if(readlink(path, dest, FILE_PATH_MAX) >= 0)
        {
            len = strlen(dest);

            for(i=len; i>0; i--)
            {
                if(dest[i] == '/')
                    break;
                else
                    dest[i] = 0;
            }

            strcat(dest, SETUP_INI);
        }
    #elif defined(__QNX__)
        char path[FILE_PATH_MAX];
        int  fd, len, i;

        memset(path, 0, sizeof(path));

        pid_t pid = getpid();

        sprintf(path, "/proc/%d/exefile", pid);

        fd = open(path, O_RDONLY);
        if(fd < 0)
        {
            perror("get execute path");
            exit(1);
        }

        len = read(fd, dest, sizeof(dest)-1);
        if(len < 0)
            exit(1);
        
        close(fd);
    
        for(i=len; i>0; i--)
        {
            if(dest[i] == '/')
                break;
            else
                dest[i] = 0;
        }

        strcat(dest, SETUP_INI);

    #else
        strcat(dest, SETUP_INI);
    #endif

    if(ini_parse(dest, handler_ini, &ini_item) < 0)
    {
        printf("Open setup.ini failed\n");

        #if defined(_WIN32)
            system("pause");
        #endif

        exit(1);
    }

} /* END: get_ini_setting() */

/*---------------------------------------------------------------------------------*/
static void show_sensor_info (int monitor_opt)
{
    int         ret           = 0;
    SENSOR_INFO sensor_info   = {0};

    /* ----- HCHO ----- */
    if(monitor_opt & SENSOR_NAME_HCHO)
    {
        printf("\n[HCHO sensor information]\n");

        sensor_info.name = SENSOR_NAME_HCHO;

        ret = UISMSensorInfo(&sensor_info);
        if(ret)
            printf("UISMSensorInfo() failed !\n");
        else
        {
            printf("UISMSensorInfo() successfully !\n");
            printf("PID        : %s\n", sensor_info.PID);
            printf("FW version : %s\n", sensor_info.FW_ver);
            printf("SN         : %s\n", sensor_info.SN);
            printf("==============================================\n");
        }
    }

    /* ----- PM ----- */
    if(monitor_opt & SENSOR_NAME_PM)
    {
        printf("\n[PM sensor information]\n");

        sensor_info.name = SENSOR_NAME_PM;

        ret = UISMSensorInfo(&sensor_info);
        if(ret)
            printf("UISMSensorInfo() failed !\n");
        else
        {
            printf("UISMSensorInfo() successfully !\n");
            printf("PID        : %s\n", sensor_info.PID);
            printf("FW version : %s\n", sensor_info.FW_ver);
            printf("SN         : %s\n", sensor_info.SN);
            printf("==============================================\n");
        }
    }

    /* ----- CO2 ----- */
    if(monitor_opt & SENSOR_NAME_CO2)
    {
        printf("\n[CO2 sensor information]\n");

        sensor_info.name = SENSOR_NAME_CO2;

        ret = UISMSensorInfo(&sensor_info);
        if(ret)
            printf("UISMSensorInfo() failed !\n");
        else
        {
            printf("UISMSensorInfo() successfully !\n");
            printf("PID        : %s\n", sensor_info.PID);
            printf("FW version : %s\n", sensor_info.FW_ver);
            printf("SN         : %s\n", sensor_info.SN);
            printf("Status     : %d\n", sensor_info.status);
            printf("==============================================\n");
        }
    }

    /* ----- CO ----- */
    if(monitor_opt & SENSOR_NAME_CO)
    {
        printf("\n[CO sensor information]\n");

        sensor_info.name = SENSOR_NAME_CO;

        ret = UISMSensorInfo(&sensor_info);
        if(ret)
            printf("UISMSensorInfo() failed !\n");
        else
        {
            printf("UISMSensorInfo() successfully !\n");
            printf("PID        : %s\n", sensor_info.PID);
            printf("FW version : %s\n", sensor_info.FW_ver);
            printf("SN         : %s\n", sensor_info.SN);
            printf("Status     : %d\n", sensor_info.status);
            printf("==============================================\n");
        }
    }

    /* ----- O3 ----- */
    if(monitor_opt & SENSOR_NAME_O3)
    {
        printf("\n[O3 sensor information]\n");

        sensor_info.name = SENSOR_NAME_O3;

        ret = UISMSensorInfo(&sensor_info);
        if(ret)
            printf("UISMSensorInfo() failed !\n");
        else
        {
            printf("UISMSensorInfo() successfully !\n");
            printf("PID        : %s\n", sensor_info.PID);
            printf("FW version : %s\n", sensor_info.FW_ver);
            printf("SN         : %s\n", sensor_info.SN);
            printf("Status     : %d\n", sensor_info.status);
            printf("==============================================\n");
        }
    }

    /* ----- TVOC ----- */
    if(monitor_opt & SENSOR_NAME_TVOC)
    {
        printf("\n[TVOC sensor information]\n");

        sensor_info.name = SENSOR_NAME_TVOC;

        ret = UISMSensorInfo(&sensor_info);
        if(ret)
            printf("UISMSensorInfo() failed !\n");
        else
        {
            printf("UISMSensorInfo() successfully !\n");
            printf("PID        : %s\n", sensor_info.PID);
            printf("FW version : %s\n", sensor_info.FW_ver);
            printf("SN         : %s\n", sensor_info.SN);
            printf("Status     : %d\n", sensor_info.status);
            printf("==============================================\n");
        }
    }

    /* ----- TH ----- */
    if(monitor_opt & SENSOR_NAME_TH)
    {
        printf("\n[TH sensor information]\n");

        sensor_info.name = SENSOR_NAME_TH;

        ret = UISMSensorInfo(&sensor_info);
        if(ret)
            printf("UISMSensorInfo() failed !\n");
        else
        {
            printf("UISMSensorInfo() successfully !\n");
            printf("SN         : %02X%02X%02X%02X\n", sensor_info.SN[0],
                                                      sensor_info.SN[1],
                                                      sensor_info.SN[2],
                                                      sensor_info.SN[3]);
            printf("==============================================\n");
        }
    }

} /* END: show_sensor_info() */

/*---------------------------------------------------------------------------------*/
/**
 * @return
 * True:  SDK receive callback is running
 * False: SDK receive callback is stop
 */
static bool demo_uism (void)
{
    bool        isRunning = false;
    int         ret       = 0;
    int         result    = 0;
    VER_INFO    ver_info  = {0};
    INIT_INFO   init_info = {0};

    /* ----- UISMDebugLevel() ----- */
    ret = UISMDebugLevel(ini_item.debug_level);
    if(ret)
        printf("UISMDebugLevel() failed ! (ret=%d)\n", ret);
    else
    {
        printf("UISMDebugLevel() successfully !\n");
        printf("==============================================\n");
    }

    /* ----- UISMOpenDevice() ----- */
    ret = UISMOpenDevice();
    if(ret)
    {
        printf("UISMOpenDevice() failed ! (ret=%d)\n", ret);

        #if defined(_WIN32)
            system("pause");
        #endif

        goto EXIT;
    }
    else
    {
        printf("UISMOpenDevice() successfully !\n");
        printf("==============================================\n");
    }

    /* ----- UISMShowVer() ----- */
    ret = UISMShowVer(&ver_info);
    if(ret)
        printf("UISMShowVer() failed !\n");
    else
    {
        printf("UISMShowVer() successfully !\n");
        printf("LIB ver: %s\n", ver_info.api);
        printf("==============================================\n");
    }

    /* ----- UISMInit() ----- */
    init_info.interval     = ini_item.interval;
    init_info.check_status = ini_item.check_status;
    init_info.data_cb      = data_cb_func;

    if(ini_item.auto_scan)
    {
        /* ----- UISMQuery() ----- */
        UISMQuery(&result);
        printf("UISMQuery() successfully !\n");
        printf("HCHO : %s\n", ((result & SENSOR_NAME_HCHO) ? "Found" : "Not Found"));
        printf("PM   : %s\n", ((result & SENSOR_NAME_PM  ) ? "Found" : "Not Found"));
        printf("CO2  : %s\n", ((result & SENSOR_NAME_CO2 ) ? "Found" : "Not Found"));
        printf("CO   : %s\n", ((result & SENSOR_NAME_CO  ) ? "Found" : "Not Found"));
        printf("O3   : %s\n", ((result & SENSOR_NAME_O3  ) ? "Found" : "Not Found"));
        printf("TVOC : %s\n", ((result & SENSOR_NAME_TVOC) ? "Found" : "Not Found"));
        printf("TH   : %s\n", ((result & SENSOR_NAME_TH  ) ? "Found" : "Not Found"));
        printf("==============================================\n");
        init_info.monitor_opt = result;
    }
    else
    {
        if(ini_item.HCHO) init_info.monitor_opt |= SENSOR_NAME_HCHO;
        if(ini_item.PM)   init_info.monitor_opt |= SENSOR_NAME_PM;
        if(ini_item.CO2)  init_info.monitor_opt |= SENSOR_NAME_CO2;
        if(ini_item.CO)   init_info.monitor_opt |= SENSOR_NAME_CO;
        if(ini_item.O3)   init_info.monitor_opt |= SENSOR_NAME_O3;
        if(ini_item.TVOC) init_info.monitor_opt |= SENSOR_NAME_TVOC;
        if(ini_item.TH)   init_info.monitor_opt |= SENSOR_NAME_TH;
    }

    if(init_info.monitor_opt == SENSOR_NAME_MONITOR_NULL)
        goto EXIT;

    /* ----- show sensor infomation ----- */
    show_sensor_info(init_info.monitor_opt);

    ret = UISMInit(INIT_STS_START, &init_info);
    if(ret)
        printf("UISMInit() failed !\n");
    else
    {
        printf("UISMInit() successfully !\n");
        printf("==============================================\n");

        isRunning = true;
    }

EXIT:

    return isRunning;

} /* END: demo_uism() */

/*---------------------------------------------------------------------------------*/
static void print_config (void)
{
    printf("\nConfig\n");
    printf("----------------------------------\n");
    printf("[setting]\n");
    printf("debug_level : %d\n",  ini_item.debug_level);
    printf("\n");
    printf("[monitor]\n");
    printf("interval     : %d\n",  ini_item.interval);
    printf("auto_scan    : %s\n", (ini_item.auto_scan    ? "True" : "False"));
    printf("check_status : %s\n", (ini_item.check_status ? "True" : "False"));
    printf("\n");
    printf("[sensor]\n");
    printf("HCHO         : %s\n", (ini_item.HCHO         ? "True" : "False"));
    printf("PM           : %s\n", (ini_item.PM           ? "True" : "False"));
    printf("CO2          : %s\n", (ini_item.CO2          ? "True" : "False"));
    printf("CO           : %s\n", (ini_item.CO           ? "True" : "False"));
    printf("O3           : %s\n", (ini_item.O3           ? "True" : "False"));
    printf("TVOC         : %s\n", (ini_item.TVOC         ? "True" : "False"));
    printf("TH           : %s\n", (ini_item.TH           ? "True" : "False"));
    printf("\n");
    printf("----------------------------------\n\n");

} /* END: print_config() */

/*---------------------------------------------------------------------------------*/
#if defined(__linux__) || defined(__QNX__)
static void end_uism (int type)
#else
static BOOL end_uism (DWORD type)
#endif
{
    #if defined(__linux__) || defined(__QNX__)
    if((type == SIGINT) || (type == TYPE_EXIT))
    #else
    if((type == CTRL_C_EVENT) || (type == TYPE_EXIT))
    #endif
    {
        printf("\nExit the program ...\n");

        UISMInit(INIT_STS_STOP, NULL);

        UISMCloseDevice();

        isRunning = false;

        if(type != TYPE_EXIT)
            exit(0);
    }

    /* prevent windows compiler warning */
    #if defined(_WIN32)
    return TRUE;
    #endif

} /* END: end_uism() */

/*---------------------------------------------------------------------------------*/
int main (void)
{
    get_ini_setting();

    print_config();

    isRunning = demo_uism();

    #if defined(__linux__) || defined(__QNX__)
        signal(SIGINT, end_uism);
    #else
        SetConsoleCtrlHandler((PHANDLER_ROUTINE) end_uism, TRUE);
    #endif

    while(isRunning)
    {
        #if defined(__linux__) || defined(__QNX__)
            usleep(100 * 1000);
        #else
            Sleep(100);
        #endif
    }

    end_uism(TYPE_EXIT);

    #if defined(_WIN32)
        getchar();
    #endif

    return 0;

} /* END: main() */
